# Global variables

id = [0,0]
matsize = [0, 0]
esp_dir = 0
 

def color16_to_32(color):
    """transform a 16 bits color (565) into a 24 bits color

    Args:
        color (int): 16 bits color in 565 format

    Returns:
        int: 24 bits color (8 bits per component)
    """

    color = int(color)
    r = ((color >> 11) & 0x1F) * 2**3
    g = ((color >> 5) & 0x3F) * 2**2
    b = (color & 0x1F) * 2**3
    return r << 16 | g << 8 | b


def disp_img_data(data):
    """Displays a buffer on the LEDs

    Args:
        data (List(int)): List of 24 bits colors to be displayed on each pixels
    """
    for i in range(NBR_LIGNES):
        for j in range(NBR_COLONNES):
            allumer_led(j, NBR_LIGNES-1-i, data[i * NBR_COLONNES + j])

def data_to_color16(data):
    """Transform the bytes received in usable 24 bits colors

    Args:
        data (bytes): Bytes received by the UART transmission

    Returns:
        List(int): List of 24 bits colors
    """
    d_color16 = []

    for i in range(0, len(data), 2):
        d_color16.append(data[i] << 8 | data[i+1])
    
    return list(map(color16_to_32, d_color16))

def display_id():
    """Display the matrix's ID on it's LEDs
    """
    afficher_texte("{}, {}".format(id[0], id[1]), speed=0.01)
 
def msg_to_str(msg):
    """Transform a byte array received via uart to a string array

    Args:
        msg (Bytes): Bytes received via uart

    Returns:
        str: Message as a string
    """
    return str(msg)[2:-1]
 
def send_passthrough_str(msg, x, y):
    """Send a passthrough command to a specific matrix, this function only work with text

    Args:
        msg (str): command to send to the x,y matrix
        x (int): x pos of the matrix
        y (int): y pos of the matrix
    """
    x = int(x)
    y = int(y)

    if x > id[0]:
        d = E
    elif x < id[0]:
        d = O
    elif y > id[1]:
        d = N
    else:
        d = S

    envoyer_msg(d, "PT;{},{};{}".format(x, y, msg))

def send_passthrough_bytes(msg, x, y):
    """Send a passthrough command to a specific matrix. This function should not be called by the user,
    it's only used when the hepialight received a PT that still needs to be transmitted 

    Args:
        msg (Bytes): Bytes received via uart
        x (int): x pos of the matrix
        y (int): y pos of the matrix
    """
    x = int(x)
    y = int(y)

    if x > id[0]:
        d = E
    elif x < id[0]:
        d = O
    elif y > id[1]:
        d = N
    else:
        d = S

    envoyer_msg(d, msg)


def send_xy(x, y, msg):
    """Send a command to a matrix in the x, y position, this function is a wrapper for send_passthrough

    Args:
        x (int): x pos of the matrix
        y (int): y pos of the matrix
        msg (str): command to send
    """
    send_passthrough_str(msg, x, y)


def send_text_xy(x, y, text, color=ROUGE, speed=0.1):
    """Tells a matrix to display text. The text will move on the matrix but not on neighbours

    Args:
        x (int): x pos of the matrix
        y (int): y pos of the matrix
        text (str): text to be displayed
        color (int, optional): 24 bits color of the text. Defaults to ROUGE.
        speed (float, optional): speed of the scrolling text. Defaults to 0.1.
    """
    send_xy(x, y, "TEXT;{};{};{}".format(color, speed, text))

def send_moving_xy(x, y, text, color=ROUGE, speed=0.1):
    """Tells a matrix to display moving text, it makes the text move on a matrix and continue
    on it's western neighbours

    Args:
        x (int): x pos of the matrix
        y (int): y pos of the matrix
        text (str): text to be displayed
        color (int, optional): 24 bits color of the text. Defaults to ROUGE.
        speed (float, optional): speed of the scrolling text. Defaults to 0.1.
    """
    send_xy(x, y, "MOVING;{};{};{}".format(color, speed, text))


def check_presence(d):
    """Check if an hepialight is present at direction d

    Args:
        d (int): Direction in which we want to check for the presence of an hepialight

    Returns:
        bool: True if an hepialight is present else False
    """
    envoyer_msg(d, "PING")

    for x in range(10):
        time.sleep(0.1)
        ret = recevoir_msg(d)

        if ret == b"PONG": return True

    return False

def update_id(new_id):
    """Update the matrix ID, and propagate the message to it's neighbours

    Args:
        new_id (List(str)): List which contains x and y coordinates ex : [x, y]
    """
    global id

    id[0] = int(new_id[0])
    id[1] = int(new_id[1])

    # The matrices which are in the x = 0 pos, update the id North and East.
    # The others only broadcast East.
    if (id[0] == 0):
        envoyer_msg(N, "SET_ID:{},{}".format(id[0], id[1] + 1))
    
    envoyer_msg(E, "SET_ID:{},{}".format(id[0] + 1, id[1]))

    if not check_presence(N) and not check_presence(E):
        send_xy(0, 0, "MATSIZE;{},{}".format(id[0], id[1]))

    display_id()



def handle_receive(msg, msg_bytes, d):
    """Handles the commands received via UART

    Args:
        msg (str): Message received as a string
        msg_bytes (bytes): Message received as a bytes
        d (int): Direction from which the message was received
    """
    global esp_dir
 
    print("Received from {} : {}".format(d, msg))

    if msg.startswith("SET_ID"):
        tmp_id = msg[7:].split(",")
        update_id(tmp_id)
 
    elif msg == "PING":
        envoyer_msg(d, "PONG")
 
    elif msg == "DISP_ID":
        display_id()

    elif msg.startswith("PT;"):
        msgs = msg.split(";")
        m = msgs[2:]
        m = ";".join(m)
        x, y = msgs[1].split(",")

        x = int(x)
        y = int(y)

        if x == id[0] and y == id[1]:
            handle_receive(m, msg_bytes[7:], d)
            return
        
        print("sending passthrough to {},{}: {}".format(x,y,m))
        send_passthrough_bytes(msg_bytes, x, y)
 
    elif msg.startswith("MATSIZE"):
        x, y = msg.split(";")[-1].split(",")
        print("Matsize : {} {}".format(x, y))
        matsize = [int(x), int(y)]
        envoyer_msg(esp_dir, "MATSIZE:{},{}".format(x, y))

    elif msg.startswith("TEXT"):
        msgs = msg.split(";")
        color = int(msgs[1])
        speed = float(msgs[2])
        afficher_texte(msgs[3], color, speed)

    elif msg.startswith("MOVING"):
        msgs 	= msg.split(";")
        color 	= int(msgs[1])
        speed 	= float(msgs[2])
        text 	= msgs[3]

        time.sleep(speed * 11)
        send_moving_xy(id[0] - 1, id[1], text, color=color, speed=speed)
        afficher_texte(text, color, speed)

    elif msg.startswith("IMAGE"):
        print("IMAGE LEN : {}".format(len(msg_bytes[6:])))
        img = data_to_color16(msg_bytes[6:])
        disp_img_data(img)

    elif msg == "OK":
        update_id(id)
        esp_dir = d


# Main
afficher_texte("Ready", VERT, speed=0.01)

# Try to find the ESP and start the initialisation
envoyer_msg(S, "ESP")
envoyer_msg(O, "ESP")

while True:

    for d in [N, S, E, O]:
        msg_bytes = recevoir_msg(d)

        if msg_bytes == b"": continue

        msg = msg_to_str(msg_bytes)
        handle_receive(msg, msg_bytes, d)